/*
 * manhattan - pu-low.c
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <compiler.h>
#include <list.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <pu-proto.h>
#include <pu-proto-lib.h>
#include <pu-lib.h>
#include <pu-buff.h>

static void pu_proto_hdr_init(struct pu_header *hdr, unsigned int id,
		unsigned int proto, unsigned int data_bytes)
{
	hdr->magic = PU_MAGIC;
	hdr->proto = proto;
	hdr->id = cpu_to_be16(id);
	hdr->data_bytes = cpu_to_be32(data_bytes);
}

struct pu_buff *pu_proto_alloc_skb(unsigned int count)
{
	struct pu_buff *n;

	count = pu_align_data(count);
	n = pu_alloc_skb(sizeof(struct pu_header) + count, GFP_KERNEL);
	if (!n)
		return NULL;
	pu_skb_reserve(n, sizeof(struct pu_header));
	return n;
}

int pu_proto_xmit_new(void *priv, struct pu_buff *skb, unsigned int proto,
		unsigned int id)
{
	struct pu_header *hdr;
	void *p;
	unsigned int padding;
	unsigned int len;

	len = skb->len;

	padding = pu_align_data(len) - len;
	if (padding) {
		p = pu_skb_put(skb, padding);
		memset(p, 0, padding);
	}

	hdr = (void *)pu_skb_push(skb, sizeof(*hdr));
	pu_proto_hdr_init(hdr, id, proto, len);
	return pu_proto_xmit(priv, skb);
}

int pu_proto_xmit_resp(void *priv, struct pu_buff *skb, struct pu_buff *req)
{
	struct pu_header *hdr;
	void *p;
	unsigned int padding;
	unsigned int len;
	unsigned int id;
	unsigned int proto;

	fprintf(stderr, "req\n");
	pu_dump_skb(req);
	hdr = (void *)pu_skb_push(req, sizeof(*hdr));
	fprintf(stderr, "x\n");
	id = be16_to_cpu(hdr->id);
	proto = hdr->proto;
	pu_consume_skb(req);

	len = skb->len;

	padding = pu_align_data(len) - len;
	if (padding) {
		p = pu_skb_put(skb, padding);
		memset(p, 0, padding);
	}

	fprintf(stderr, "skb_push\n");
	pu_dump_skb(skb);
	hdr = (void *)pu_skb_push(skb, sizeof(*hdr));
	fprintf(stderr, "skb_push\n");
	pu_proto_hdr_init(hdr, id, proto, len);

	return pu_proto_xmit(priv, skb);
}

int pu_proto_error(void *priv, struct pu_buff *req, unsigned int error)
{
	struct pu_header *hdr;
	struct pu_buff *skb = pu_alloc_skb(sizeof(*hdr), GFP_KERNEL);
	unsigned int id;
	unsigned int proto;

	hdr = (void *) pu_skb_push(req, sizeof(*hdr));
	id = be16_to_cpu(hdr->id);
	proto = hdr->proto;
	pu_consume_skb(req);
	fprintf(stderr, "id: %d\n", id);

	hdr = (void *) pu_skb_put(skb, sizeof(*hdr));
	pu_proto_hdr_init(hdr, id, proto, PU_DATA_ERROR | error);
	return pu_proto_xmit(priv, skb);
}

int pu_proto_error_errno(void *priv, struct pu_buff *req, unsigned int error)
{
	unsigned int err = PU_EUNKNOWN;

	switch (error) {
	case ENOENT:
		err = PU_ENOENT;
		break;
	case ENXIO:
		err = PU_ENXIO;
		break;
	case EAGAIN:
		err = PU_EAGAIN;
		break;
	case EINVAL:
		err = PU_EINVAL;
		break;
	case ENOSYS:
		err = PU_ENOSYS;
		break;
	}
	return pu_proto_error(priv, req, err);
}
